import asyncio
import sys
import re
import contextlib
import datetime
from pathlib import Path
from typing import Union
from bilireq.utils import get

import httpx
import nonebot
from nonebot import on_command as _on_command
from nonebot import require
from nonebot.adapters.red import (
    Bot,
    Message,
    MessageEvent,
    MessageSegment,
)
from nonebot.adapters.red.event import GroupMessageEvent, PrivateMessageEvent
from nonebot.adapters.red.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.adapters.red.api.model import ChatType
from nonebot.exception import FinishedException
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import ArgPlainText, CommandArg, RawCommand
from nonebot.permission import SUPERUSER, Permission
from nonebot.rule import Rule

from ..config import plugin_config

require("nonebot_plugin_guild_patch")
from nonebot_plugin_guild_patch import ChannelDestroyedNoticeEvent, GuildMessageEvent  # noqa


def get_path(*other):
    """获取数据文件绝对路径"""
    if plugin_config.haruka_dir:
        dir_path = Path(plugin_config.haruka_dir).resolve()
    else:
        dir_path = Path.cwd().joinpath("data")
        # dir_path = Path.cwd().joinpath('data', 'haruka_bot')
    return str(dir_path.joinpath(*other))


async def handle_uid(
    matcher: Matcher,
    command_arg: Message = CommandArg(),
):
    if command_arg.extract_plain_text().strip():
        matcher.set_arg("uid", command_arg)


async def uid_check(
    matcher: Matcher,
    uid: str = ArgPlainText("uid"),
):
    uid = uid.strip()
    if extract := await uid_extract(uid):
        uid = extract
    else:
        await matcher.finish("未找到该 UP，请输入正确的 UP 群内昵称、UP 名、UP UID或 UP 首页链接")
    matcher.set_arg("uid", Message(uid))


async def b23_extract(text: str):
    if "b23.tv" not in text and "b23.wtf" not in text:
        return None
    if not (b23 := re.compile(r"b23.(tv|wtf)[\\/]+(\w+)").search(text)):
        return None
    try:
        url = f"https://b23.tv/{b23[2]}"
        for _ in range(3):
            with contextlib.suppress(Exception):
                resp = await httpx.AsyncClient().get(url, follow_redirects=True)
                break
        else:
            return None
        url = resp.url
        logger.debug(f"b23.tv url: {url}")
        return str(url)
    except TypeError:
        return None


async def search_user(keyword: str):
    """
    搜索用户
    """
    url = "https://api.bilibili.com/x/web-interface/search/type"
    data = {"keyword": keyword, "search_type": "bili_user"}
    resp = await get(url, params=data)
    logger.debug(resp)
    return resp


async def uid_extract(text: str):
    logger.debug(f"[UID Extract] Original Text: {text}")
    b23_msg = await b23_extract(text) if "b23.tv" in text else None
    message = b23_msg or text
    logger.debug(f"[UID Extract] b23 extract: {message}")
    pattern = re.compile("^[0-9]*$|bilibili.com/([0-9]*)")
    if match := pattern.search(message):
        logger.debug(f"[UID Extract] Digit or Url: {match}")
        match = match[1] or match[0]
        return str(match)
    elif message.startswith("UID:"):
        pattern = re.compile("^\\d+")
        if match := pattern.search(message[4:]):
            logger.debug(f"[UID Extract] UID: {match}")
            return str(match[0])
    else:
        text_u = text.strip(""""'“”‘’""")
        if text_u != text:
            logger.debug(f"[UID Extract] Text is a Quoted Digit: {text_u}")
        logger.debug(f"[UID Extract] Searching UID in BiliBili: {text_u}")
        resp = await search_user(text_u)
        logger.debug(f"[UID Extract] Search result: {resp}")
        if resp and resp["numResults"]:
            for result in resp["result"]:
                if result["uname"] == text_u:
                    logger.debug(f"[UID Extract] Found User: {result}")
                    return str(result["mid"])
        logger.debug("[UID Extract] No User found")


async def _guild_admin(bot: Bot, event: GuildMessageEvent):
    roles = {
        role["role_name"]
        for role in (
            await bot.get_guild_member_profile(guild_id=event.guild_id, user_id=event.user_id)
        )["roles"]
    }
    return bool(roles & set(plugin_config.haruka_guild_admin_roles))


GUILD_ADMIN: Permission = Permission(_guild_admin)


async def permission_check(
    bot: Bot, event: Union[GroupMessageEvent, PrivateMessageEvent, GuildMessageEvent]
):
    from ..database import DB as db

    if isinstance(event, PrivateMessageEvent):
        if event.subMsgType == -1:  # 占位，无实际作用
            raise FinishedException
        return
    if isinstance(event, GroupMessageEvent):
        if not await db.get_group_admin(event.group_id):
            return
        if await (GROUP_ADMIN | GROUP_OWNER | SUPERUSER)(bot, event):
            return
    elif isinstance(event, GuildMessageEvent):
        if not await db.get_guild_admin(event.guild_id, event.channel_id):
            return
        if await (GUILD_ADMIN | SUPERUSER)(bot, event):
            return
    await bot.send(event, "权限不足，目前只有管理员才能使用")
    raise FinishedException


async def group_only(matcher: Matcher, event: PrivateMessageEvent, command: str = RawCommand()):
    await matcher.finish(f"只有群里才能{command}")


def to_me():
    if plugin_config.haruka_to_me:
        from nonebot.rule import to_me

        return to_me()

    async def _to_me() -> bool:
        return True

    return Rule(_to_me)


def calc_time_total(t):
    t = int(t * 1000)
    if t < 5000:
        return f"{t} 毫秒"

    timedelta = datetime.timedelta(seconds=t // 1000)
    day = timedelta.days
    hour, mint, sec = tuple(int(n) for n in str(timedelta).split(",")[-1].split(":"))

    total = ""
    if day:
        total += f"{day} 天 "
    if hour:
        total += f"{hour} 小时 "
    if mint:
        total += f"{mint} 分钟 "
    if sec and not day and not hour:
        total += f"{sec} 秒 "
    return total


async def safe_send(bot_id, send_type, type_id, message, at=False):
    """发送出现错误时, 尝试重新发送, 并捕获异常且不会中断运行"""

    async def _safe_send(bot: Bot, send_type, type_id, message):
        if send_type == "guild":
            # from ..database import DB as db
            #
            # guild = await db.get_guild(id=type_id)
            # assert guild
            # result = await bot.send_guild_channel_msg(
            #     guild_id=guild.guild_id,
            #     channel_id=guild.channel_id,
            #     message=message,
            # )
            return   # 暂时不处理频道消息
        else:
            if send_type == "group":
                send_type = ChatType.GROUP
            else:
                send_type = ChatType.FRIEND
            result = await bot.send_message(
                send_type,
                message=message,
                target=type_id,
            )
        return result

    bots = nonebot.get_bots()
    bot = bots.get(str(bot_id))
    if bot is None:
        logger.error(f"推送失败，Bot（{bot_id}）未连接，尝试使用其他 Bot 推送")
        for bot_id, bot in bots.items():
            if at and (
                send_type == "guild"
                or (await bot.get_group_at_all_remain(group_id=type_id))["can_at_all"]
            ):
                message = MessageSegment.at("all") + message
            try:
                result = await _safe_send(bot, send_type, type_id, message)
                logger.info(f"尝试使用 Bot（{bot_id}）推送成功")
                return result
            except Exception:
                continue
        logger.error("尝试失败，所有 Bot 均无法推送")
        return

    # if at and (
    #     send_type == "guild"
    #     or (await bot.get_group_at_all_remain(group_id=type_id))["can_at_all"]
    # ):
    if at:
        message = MessageSegment.at_all() + " " + message

    try:
        return await _safe_send(bot, send_type, type_id, message)
    except Exception as e:
        logger.error(f"推送失败，错误信息：{e}")


async def get_type_id(event: Union[MessageEvent, ChannelDestroyedNoticeEvent]):
    if isinstance(event, (GuildMessageEvent, ChannelDestroyedNoticeEvent)):
        from ..database import DB as db

        return await db.get_guild_type_id(event.guild_id, event.channel_id)
    return event.group_id if isinstance(event, GroupMessageEvent) else event.user_id


def check_proxy():
    """检查代理是否有效"""
    if plugin_config.haruka_proxy:
        logger.info("检查代理是否有效")
        try:
            httpx.get(
                "https://icanhazip.com/",
                proxies={"all://": plugin_config.haruka_proxy},
                timeout=2,
            )
        except Exception:
            raise RuntimeError("加载失败，代理无法连接，请检查 HARUKA_PROXY 后重试")


def on_startup():
    """安装依赖并检查当前环境是否满足运行条件"""
    if plugin_config.fastapi_reload and sys.platform == "win32":
        raise ImportError("加载失败，Windows 必须设置 FASTAPI_RELOAD=false 才能正常运行 HarukaBot")
    try:  # 如果开启 realod 只在第一次运行
        asyncio.get_running_loop()
    except RuntimeError:
        from .browser import check_playwright_env, install

        check_proxy()
        install()
        asyncio.get_event_loop().run_until_complete(check_playwright_env())
        # 创建数据存储目录
        if not Path(get_path()).is_dir():
            Path(get_path()).mkdir(parents=True)


def on_command(cmd, *args, **kwargs):
    return _on_command(plugin_config.haruka_command_prefix + cmd, *args, **kwargs)


PROXIES = {"all://": plugin_config.haruka_proxy}

require("nonebot_plugin_apscheduler")
from nonebot_plugin_apscheduler import scheduler  # noqa

from .browser import get_dynamic_screenshot  # noqa
from .get_dynamic_list import get_user_dynamics  # noqa



from typing import Optional

from pydantic import BaseModel


class Info(BaseModel):
    """用户信息"""

    uid: Optional[int]
    uname: Optional[str]
    face: Optional[str]
    head_url: Optional[str]
    name: Optional[str]


class LevelInfo(BaseModel):
    """等级信息"""

    current_level: Optional[int]


class Pendant(BaseModel):
    """挂件"""

    pid: int
    name: str
    image: str


class OfficialVerify(BaseModel):
    """账号认证信息"""

    type: int
    desc: str


class Card(BaseModel):
    official_verify: OfficialVerify


class VIP(BaseModel):
    """大会员信息"""

    vipType: int
    nickname_color: str


class UserProfile(BaseModel):
    info: Info
    level_info: Optional[LevelInfo]
    pendant: Optional[Pendant]
    card: Optional[Card]
    vip: Optional[VIP]



import asyncio
import contextlib
from io import BytesIO
import os
import re
import sys
from pathlib import Path
from typing import Optional

import skia
from nonebot.log import logger
# from aunly_captcha_solver import CaptchaInfer
from playwright.__main__ import main
from playwright.async_api import BrowserContext, async_playwright, Page
from bilireq.utils import get, DEFAULT_HEADERS
from dynrender_skia.Core import DynRender
from dynamicadaptor.DynamicConversion import formate_message

from ..config import plugin_config
from .fonts_provider import fill_font
from ..utils import get_path

_browser: Optional[BrowserContext] = None
mobile_js = Path(__file__).parent.joinpath("mobile.js")


async def init_browser(proxy=plugin_config.haruka_proxy, **kwargs) -> BrowserContext:
    logger.info("初始化浏览器")
    if proxy:
        kwargs["proxy"] = {"server": proxy}
    global _browser
    p = await async_playwright().start()
    browser_data = Path(get_path("browser"))
    browser_data.mkdir(parents=True, exist_ok=True)
    browser_context = await p.chromium.launch_persistent_context(
        browser_data,
        user_agent=plugin_config.haruka_browser_ua
        or (
            (
                "Mozilla/5.0 (Linux; Android 10; RMX1911) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/113.0.0.0 Mobile Safari/537.36"
            )
            if plugin_config.haruka_screenshot_style.lower() == "mobile"
            else None
        ),
        device_scale_factor=2,
        timeout=plugin_config.haruka_dynamic_timeout * 1000,
        **kwargs,
    )
    if plugin_config.haruka_screenshot_style.lower() != "mobile":
        await browser_context.add_cookies(
            [
                {
                    "name": "hit-dyn-v2",
                    "value": "1",
                    "domain": ".bilibili.com",
                    "path": "/",
                }
            ]
        )
    _browser = browser_context
    return _browser


async def get_browser() -> BrowserContext:
    global _browser
    if not _browser:
        _browser = await init_browser()
    return _browser


async def get_dynamic_screenshot(dynamic_id, style=plugin_config.haruka_screenshot_style):
    """获取动态截图"""
    image: Optional[bytes] = None
    err = ""
    if style == "local":
        return (await get_dynamic_screenshot_local(dynamic_id), err)
    else:
        for i in range(3):
            browser = await get_browser()
            page = await browser.new_page()
            try:
                # if style.lower() == "mobile":
                #     page, clip = await get_dynamic_screenshot_mobile(dynamic_id, page)
                # else:
                #     page, clip = await get_dynamic_screenshot_pc(dynamic_id, page)
                page, clip = await get_dynamic_screenshot_mobile(dynamic_id, page)
                clip["height"] = min(clip["height"], 32766)
                return (
                    await page.screenshot(clip=clip, full_page=True, type="jpeg", quality=98),
                    None,
                )
            except TimeoutError:
                logger.warning(f"截图超时，重试 {i + 1}/3")
                err = "截图超时"
            except Notfound:
                logger.error(f"动态 {dynamic_id} 不存在")
                err = "动态不存在"
            except AssertionError:
                logger.error(f"动态 {dynamic_id} 截图失败")
                err = "网页元素获取失败"
                image = await get_dynamic_screenshot_local(dynamic_id)
            except Exception as e:
                if "bilibili.com/404" in page.url:
                    logger.error(f"动态 {dynamic_id} 不存在")
                    err = "动态不存在"
                    break
                elif "waiting until" in str(e):
                    logger.error(f"动态 {dynamic_id} 截图超时")
                    err = "截图超时"
                else:
                    logger.exception(f"动态 {dynamic_id} 截图失败")
                    err = "截图失败"
                    with contextlib.suppress(Exception):
                        image = await get_dynamic_screenshot_local(dynamic_id)
            finally:
                with contextlib.suppress(Exception):
                    await page.close()
        return image, err
    

async def get_dynamic_screenshot_local(dynamic_id) -> bytes:
    """本地渲染动态截图"""
    url = f"https://api.bilibili.com/x/polymer/web-dynamic/v1/detail?timezone_offset=-480&id={dynamic_id}&features=itemOpusStyle"
    headers = {
        "referer": f"https://t.bilibili.com/{dynamic_id}",
        "user-agent": plugin_config.haruka_browser_ua,
        "Cookie": plugin_config.haruka_browser_cookie,
    }
    message = await get(url, headers={**DEFAULT_HEADERS, **headers})

    message_formate = await formate_message("web", message["item"])
    img = await DynRender(font_family=plugin_config.haruka_dynamic_font).run(message_formate)
    img = skia.Image.fromarray(img, colorType=skia.ColorType.kRGBA_8888_ColorType)
    bytes_io = BytesIO()
    img.save(bytes_io)
    return bytes_io.getvalue()


async def get_dynamic_screenshot_mobile(dynamic_id, page: Page):
    """移动端动态截图"""
    url = f"https://m.bilibili.com/dynamic/{dynamic_id}"
    await page.set_viewport_size({"width": 460, "height": 780})
    await page.route(re.compile("^https://static.graiax/fonts/(.+)$"), fill_font)
    # if plugin_config.haruka_captcha_address:
    #     captcha = CaptchaInfer(
    #         plugin_config.haruka_captcha_address, plugin_config.haruka_captcha_token
    #     )
    #     page = await captcha.solve_captcha(page, url)
    # else:
    await page.goto(url, wait_until="networkidle")
    # 动态被删除或者进审核了
    if page.url == "https://m.bilibili.com/404":
        raise Notfound
    # await page.add_script_tag(
    #     content=
    #     # 去除打开app按钮
    #     "document.getElementsByClassName('m-dynamic-float-openapp').forEach(v=>v.remove());"
    #     # 去除关注按钮
    #     "document.getElementsByClassName('dyn-header__following').forEach(v=>v.remove());"
    #     # 修复字体与换行问题
    #     "const dyn=document.getElementsByClassName('dyn-card')[0];"
    #     "dyn.style.fontFamily='Noto Sans CJK SC, sans-serif';"
    #     "dyn.style.overflowWrap='break-word'"
    # )

    await page.wait_for_load_state(state="domcontentloaded")
    await page.wait_for_selector(".b-img__inner, .dyn-header__author__face", state="visible")

    await page.add_script_tag(path=mobile_js)

    await page.evaluate(
        f'setFont("{plugin_config.haruka_dynamic_font}", '
        f'"{plugin_config.haruka_dynamic_font_source}")'
        if plugin_config.haruka_dynamic_font
        else "setFont()"
    )
    await page.wait_for_function(
        f"getMobileStyle({'true' if plugin_config.haruka_dynamic_big_image else 'false'})"
    )

    await page.wait_for_load_state("networkidle")
    await page.wait_for_load_state("domcontentloaded")

    await page.wait_for_timeout(
        1000 if plugin_config.haruka_dynamic_font_source == "remote" else 200
    )

    # 判断字体是否加载完成
    need_wait = ["imageComplete", "fontsLoaded"]
    await asyncio.gather(*[page.wait_for_function(f"{i}()") for i in need_wait])

    card = await page.query_selector(".opus-modules" if "opus" in page.url else ".dyn-card")
    assert card
    clip = await card.bounding_box()
    assert clip
    return page, clip


async def get_dynamic_screenshot_pc(dynamic_id, page: Page):
    """电脑端动态截图"""
    url = f"https://t.bilibili.com/{dynamic_id}"
    await page.set_viewport_size({"width": 2560, "height": 1080})
    await page.goto(url, wait_until="networkidle")
    # 动态被删除或者进审核了
    if page.url == "https://www.bilibili.com/404":
        raise Notfound
    card = await page.query_selector(".card")
    assert card
    clip = await card.bounding_box()
    assert clip
    bar = await page.query_selector(".bili-dyn-action__icon")
    assert bar
    bar_bound = await bar.bounding_box()
    assert bar_bound
    clip["height"] = bar_bound["y"] - clip["y"]
    return page, clip


def install():
    """自动安装、更新 Chromium"""

    def restore_env():
        del os.environ["PLAYWRIGHT_DOWNLOAD_HOST"]
        if plugin_config.haruka_proxy:
            del os.environ["HTTPS_PROXY"]
        if original_proxy is not None:
            os.environ["HTTPS_PROXY"] = original_proxy

    logger.info("检查 Chromium 更新")
    sys.argv = ["", "install", "chromium"]
    original_proxy = os.environ.get("HTTPS_PROXY")
    if plugin_config.haruka_proxy:
        os.environ["HTTPS_PROXY"] = plugin_config.haruka_proxy
    os.environ["PLAYWRIGHT_DOWNLOAD_HOST"] = "https://npmmirror.com/mirrors/playwright/"
    success = False
    try:
        main()
    except SystemExit as e:
        if e.code == 0:
            success = True
    if not success:
        logger.info("Chromium 更新失败，尝试从原始仓库下载，速度较慢")
        os.environ["PLAYWRIGHT_DOWNLOAD_HOST"] = ""
        try:
            main()
        except SystemExit as e:
            if e.code != 0:
                restore_env()
                raise RuntimeError("未知错误，Chromium 下载失败")
    restore_env()


async def check_playwright_env():
    """检查 Playwright 依赖"""
    logger.info("检查 Playwright 依赖")
    try:
        async with async_playwright() as p:
            await p.chromium.launch()
    except Exception:
        raise ImportError(
            "加载失败，Playwright 依赖不全，" "解决方法：https://haruka-bot.sk415.icu/faq.html#playwright-依赖不全"
        )


class Notfound(Exception):
    pass



from typing import Union

from nonebot.adapters.red.event import GroupMessageEvent
from nonebot.adapters.red.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.permission import SUPERUSER
from nonebot_plugin_guild_patch import GuildMessageEvent

from ...database import DB as db
from ...utils import GUILD_ADMIN, group_only, on_command, to_me

permission_off = on_command(
    "关闭权限",
    rule=to_me(),
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER | GUILD_ADMIN,
    priority=5,
)
permission_off.__doc__ = """关闭权限"""

permission_off.handle()(group_only)


@permission_off.handle()
async def _(event: Union[GroupMessageEvent, GuildMessageEvent]):
    """关闭当前群权限"""
    if isinstance(event, GuildMessageEvent):
        if await db.set_guild_permission(event.guild_id, event.channel_id, False):
            await permission_off.finish("已开启权限，只有管理员和主人可以操作")
    else:
        if await db.set_permission(event.group_id, False):
            await permission_off.finish("已关闭权限，所有人均可操作")
    await permission_off.finish("权限已经关闭了，所有人均可操作")



from pathlib import Path
from io import BytesIO
import time

import httpx
from bilireq.live import get_rooms_info_by_uids
from nonebot.adapters.red.message import MessageSegment
from nonebot.log import logger

from ...config import plugin_config
from ...database import DB as db
from ...utils import PROXIES, safe_send, scheduler, calc_time_total

status = {}
live_time = {}


@scheduler.scheduled_job(
    "interval", seconds=plugin_config.haruka_live_interval, id="live_sched"
)
async def live_sched():
    # sourcery skip: use-fstring-for-concatenation
    """直播推送"""
    uids = await db.get_uid_list("live")

    if not uids:  # 订阅为空
        return
    logger.debug(f"爬取直播列表，目前开播{sum(status.values())}人，总共{len(uids)}人")
    res = await get_rooms_info_by_uids(uids, reqtype="web", proxies=PROXIES)
    if not res:
        return
    for uid, info in res.items():
        new_status = 0 if info["live_status"] == 2 else info["live_status"]
        if uid not in status:
            status[uid] = new_status
            continue
        old_status = status[uid]
        if new_status == old_status:  # 直播间状态无变化
            continue
        status[uid] = new_status

        name = info["uname"]
        if new_status:  # 开播
            live_time[uid] = info["live_time"]
            room_id = info["short_id"] or info["room_id"]
            url = f"https://live.bilibili.com/{room_id}"
            title = info["title"]
            cover = info["cover_from_user"] or info["keyframe"]
            cover_bytesio = BytesIO(httpx.get(cover).content)
            area = info["area_v2_name"]
            area_parent = info["area_v2_parent_name"]
            room_area = f"{area_parent} / {area}"
            logger.info(f"检测到开播：{name}（{uid}）")
            live_msg = (
                f"{name} 开播啦！\n分区：{room_area}\n标题：{title}\n"
                + MessageSegment.image(cover_bytesio)
                + f"\n{url}"
            )
        else:  # 下播
            logger.info(f"检测到下播：{name}（{uid}）")
            if not plugin_config.haruka_live_off_notify:  # 没开下播推送
                continue
            live_time_msg = (
                f"\n本次直播时长 {calc_time_total(time.time() - live_time[uid])}。"
                if live_time.get(uid)
                else "。"
            )
            live_msg = f"{name} 下播了{live_time_msg}"

        # 推送
        push_list = await db.get_push_list(uid, "live")
        for sets in push_list:
            await safe_send(
                bot_id=sets.bot_id,
                send_type=sets.type,
                type_id=sets.type_id,
                message=live_msg,
                at=bool(sets.at) if new_status else False,  # 下播不@全体
            )
        await db.update_user(int(uid), name)



from playwright.async_api import Request, Route
from yarl import URL
from loguru import logger

from pathlib import Path

font_path = Path("font")

font_mime_map = {
    "collection": "font/collection",
    "otf": "font/otf",
    "sfnt": "font/sfnt",
    "ttf": "font/ttf",
    "woff": "font/woff",
    "woff2": "font/woff2",
}


async def fill_font(route: Route, request: Request):
    url = URL(request.url)
    if font_path.joinpath(url.name).exists():
        logger.debug(f"Font {url.name} found in local")
        await route.fulfill(
            path=font_path.joinpath(url.name),
            content_type=font_mime_map.get(url.suffix, None),
        )
        return
    await route.fallback()



from nonebot.plugin import PluginMetadata
from nonebot.plugin.manager import PluginLoader

if isinstance(globals()["__loader__"], PluginLoader):
    from .utils import on_startup

    on_startup()

    from . import plugins  # noqa: F401

from .version import VERSION, __version__  # noqa: F401

__plugin_meta__ = PluginMetadata(
    name="haruka_bot_red",
    description="将B站UP主的动态和直播信息推送至QQ，适配red协议",
    usage="https://github.com/boxie123/HarukaBot-Red",
    homepage="https://github.com/boxie123/HarukaBot-Red",
    type="application",
    supported_adapters={"~red"},
    extra={
        "author": "boxie123",
        "version": __version__,
        "priority": 1,
    },
)



from . import help  # noqa: F401
from .at import at_off, at_on  # noqa: F401
from .dynamic import dynamic_off, dynamic_on  # noqa: F401
from .permission import permission_off, permission_on  # noqa: F401
from .pusher import dynamic_pusher, live_pusher  # noqa: F401
from .live import live_off, live_on, live_now  # noqa: F401
from .sub import add_sub, delete_sub, sub_list  # noqa: F401



from nonebot.matcher import matchers

from ..utils import on_command, to_me, permission_check
from ..version import __version__

help = on_command("帮助", rule=to_me(), priority=5)

help.handle()(permission_check)


@help.handle()
async def _():
    message = "目前支持的功能：\n"
    for matchers_list in matchers.values():
        for matcher in matchers_list:
            if (
                matcher.plugin_name
                and matcher.plugin_name.startswith("haruka_bot")
                and matcher.__doc__
            ):
                message += matcher.__doc__ + "\n"
    message += (
        f"\n当前版本：v{__version__}\n"
    )
    await help.finish(message)



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Json


class TopicDetail(BaseModel):
    topic_name: str
    is_activity: bool


class TopicInfo(BaseModel):
    topic_details: List[TopicDetail]


class EmojiDetail(BaseModel):
    emoji_name: str
    id: int
    text: str
    url: str


class EmojiInfo(BaseModel):
    """emoji 信息"""

    emoji_details: List[EmojiDetail]


class DescFirst(BaseModel):
    text: str


class ReserveAttachCard(BaseModel):
    title: Optional[str]
    desc_first: Optional[Union[str, DescFirst]]
    desc_second: Optional[str]
    cover_url: Optional[str]
    head_text: Optional[str]


class AddOnCardInfo(BaseModel):
    add_on_card_show_type: int
    reserve_attach_card: Optional[ReserveAttachCard]
    vote_card: Optional[Json]
    attach_card: Optional[ReserveAttachCard]


class Display(BaseModel):
    topic_info: Optional[TopicInfo]
    emoji_info: Optional[EmojiInfo]
    add_on_card_info: Optional[List[AddOnCardInfo]]
    origin: Optional[Dict]



from nonebot.adapters.red.event import MessageEvent
from nonebot.params import ArgPlainText

from ...database import DB as db
from ...utils import (
    get_type_id,
    handle_uid,
    on_command,
    permission_check,
    to_me,
    uid_check,
)

live_off = on_command("关闭直播", rule=to_me(), priority=5)
live_off.__doc__ = """关闭直播 UID"""

live_off.handle()(permission_check)

live_off.handle()(handle_uid)

live_off.got("uid", prompt="请输入要关闭直播的UID")(uid_check)


@live_off.handle()
async def _(event: MessageEvent, uid: str = ArgPlainText("uid")):
    """根据 UID 关闭直播"""
    if event.chatType == 2:
        message_type = "group"
    else:
        message_type = "private"
    if await db.set_sub(
        "live",
        False,
        uid=uid,
        type=message_type,
        type_id=await get_type_id(event),
    ):
        user = await db.get_user(uid=uid)
        assert user is not None
        await live_off.finish(f"已关闭 {user.name}（{user.uid}）的直播推送")
    await live_off.finish(f"UID（{uid}）未关注，请先关注后再操作")



from bilireq.utils import get, DEFAULT_HEADERS


async def get_user_dynamics(
        uid: int, cookie: str, ua: str, **kwargs
):
    """根据 UID 获取动态列表"""
    url = "https://api.bilibili.com/x/polymer/web-dynamic/v1/feed/space"
    data = {"host_mid": uid}
    headers = {
        **DEFAULT_HEADERS,
        **{"Cookie": cookie,
           "User-Agent": ua,
           "Origin": "https://space.bilibili.com",
           "Referer": f"https://space.bilibili.com/{uid}/dynamic"}}
    return await get(url, params=data, headers=headers)



from typing import Optional

from pydantic import BaseModel

from .user_profile import UserProfile


class Desc(BaseModel):
    type: int
    timestamp: int
    view: int
    orig_dy_id: Optional[int]
    orig_type: int
    user_profile: UserProfile
    dynamic_id: int
    # uid: int
    # rid: int
    # acl: int
    # repost: int
    # comment: int
    # like: int
    # is_liked: int
    # pre_dy_id: int
    # uid_type: int
    # stype: int
    # r_type: int
    # inner_id: int
    # status: int
    # dynamic_id_str: str
    # pre_dy_id_str: str
    # orig_dy_id_str: str
    # rid_str: str



import asyncio
from datetime import datetime

from apscheduler.events import (
    EVENT_JOB_ERROR,
    EVENT_JOB_EXECUTED,
    EVENT_JOB_MISSED,
    EVENT_SCHEDULER_STARTED,
)
from nonebot.adapters.red.message import MessageSegment
from nonebot.log import logger

from ...config import plugin_config
from ...database import DB as db
from ...database import dynamic_offset as offset
from ...utils import get_dynamic_screenshot, safe_send, scheduler, get_user_dynamics


async def dy_sched():
    """动态推送"""
    uid = await db.next_uid("dynamic")
    if not uid:
        # 没有订阅先暂停一秒再跳过，不然会导致 CPU 占用过高
        await asyncio.sleep(1)
        return
    user = await db.get_user(uid=uid)
    assert user is not None
    name = user.name

    logger.debug(f"爬取动态 {name}（{uid}）")
    # 获取 UP 最新动态列表
    dynamics: list = (
        await get_user_dynamics(
            uid,
            cookie=plugin_config.haruka_browser_cookie,
            ua=plugin_config.haruka_browser_ua,
            timeout=plugin_config.haruka_dynamic_timeout,
            proxy=plugin_config.haruka_proxy,
        )
    )["items"]

    if not dynamics:  # 没发过动态
        if uid in offset and offset[uid] == [-1, ]:  # 不记录会导致第一次发动态不推送
            offset[uid].append(0)
        return
    # 更新昵称
    name = dynamics[0]["modules"]["module_author"]["name"]

    dynamics.sort(key=lambda x: int(x["id_str"]))  # 动态从旧到新排列

    if uid not in offset:  # 已删除
        return
    elif offset[uid] == [-1, ]:  # 第一次爬取
        if len(dynamics) == 1:  # 只有一条动态
            offset[uid].append(int(dynamics[0]["id_str"]))
        else:
            offset[uid] = [int(x["id_str"]) for x in dynamics]  # 记录前12条动态id
        return

    dynamic = None
    for dynamic in dynamics:
        dynamic_id = int(dynamic["id_str"])

        if (dynamic_id not in offset[uid]) and (dynamic_id > offset[uid][1]):  # 和记录中第二旧的动态比较，排除置顶变动影响
            logger.info(f"检测到新动态（{dynamic_id}）：{name}（{uid}）")
            url = f"https://t.bilibili.com/{dynamic_id}"
            if dynamic["type"] in [
                "DYNAMIC_TYPE_LIVE_RCMD",
                "DYNAMIC_TYPE_LIVE",
                "DYNAMIC_TYPE_AD",
                "DYNAMIC_TYPE_BANNER",
            ]:
                logger.debug(f"无需推送的动态 {dynamic['type']}，已跳过：{url}")
                offset[uid].append(dynamic_id)
                continue
            image, err = await get_dynamic_screenshot(dynamic_id)
            if image is None:
                logger.debug(f"动态不存在，已跳过：{url}")
                continue

            type_msg = {
                0: "发布了新动态",
                "DYNAMIC_TYPE_FORWARD": "转发了一条动态",
                "DYNAMIC_TYPE_WORD": "发布了新文字动态",
                "DYNAMIC_TYPE_DRAW": "发布了新图文动态",
                "DYNAMIC_TYPE_AV": "发布了新投稿",
                "DYNAMIC_TYPE_ARTICLE": "发布了新专栏",
                "DYNAMIC_TYPE_MUSIC": "发布了新音频",
            }

            message = (
                f"{name} {type_msg.get(dynamic['type'], type_msg[0])}：\n"
                + str(f"动态图片可能截图异常：{err}\n" if err else "")
                + MessageSegment.image(image)
                + f"\n{url}"
            )

            push_list = await db.get_push_list(uid, "dynamic")
            for sets in push_list:
                await safe_send(
                    bot_id=sets.bot_id,
                    send_type=sets.type,
                    type_id=sets.type_id,
                    message=message,
                    at=bool(sets.at) and plugin_config.haruka_dynamic_at,
                )

            offset[uid].append(dynamic_id)
            offset[uid].sort()

    if dynamic:
        await db.update_user(uid, name)

    if len(offset[uid]) > 12:  # 删除已不在前12条的动态id，防止记录过多拖慢运行速度
        offset[uid] = offset[uid][1:]


def dynamic_lisener(event):
    if hasattr(event, "job_id") and event.job_id != "dynamic_sched":
        return
    job = scheduler.get_job("dynamic_sched")
    if not job:
        scheduler.add_job(
            dy_sched, id="dynamic_sched", next_run_time=datetime.now(scheduler.timezone)
        )


if plugin_config.haruka_dynamic_interval == 0:
    scheduler.add_listener(
        dynamic_lisener,
        EVENT_JOB_EXECUTED | EVENT_JOB_ERROR | EVENT_JOB_MISSED | EVENT_SCHEDULER_STARTED,
    )
else:
    scheduler.add_job(
        dy_sched,
        "interval",
        seconds=plugin_config.haruka_dynamic_interval,
        id="dynamic_sched",
    )



from nonebot.adapters.red.event import MessageEvent

from ...database import DB as db
from ...utils import get_type_id, on_command, permission_check, to_me

from ..pusher.live_pusher import status

live_now = on_command("已开播", rule=to_me(), priority=5)
live_now.__doc__ = """已开播"""

live_now.handle()(permission_check)


@live_now.handle()
async def _(event: MessageEvent):
    """返回已开播的直播间"""
    if event.chatType == 2:
        message_type = "group"
    else:
        message_type = "private"
    subs = await db.get_sub_list(message_type, await get_type_id(event))
    if now_live := [sub for sub in subs if status.get(str(sub.uid)) == 1]:
        await live_now.finish(
            f"共有{len(now_live)}个主播正在直播：\n\n"
            + "\n".join([f"{await db.get_name(sub.uid)}（{sub.uid}）" for sub in now_live])
        )
    await live_now.finish("当前没有正在直播的主播")
